#Neural Network Functions
#Robot Personality Simulator
'''File contains constructors and learning update functions
    for various types of neural networks'''

'''created by Casey Bennett 2019, www.CaseyBennett.com
   Copyright 2019

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License (Lesser GPL) as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
'''

import os
import sys
import math
import numpy as np

from keras.models import Sequential, Model
from keras.layers import Input, Dense, Activation, Dropout, concatenate, SimpleRNN, LSTM

import cfg


#############################################################################
#
# Basic MLP Neural Network
# 
##########################

def MLP_construct():
    '''Constructor for Basic MLP feed forward neural network'''

    #Create Input Layers
    shape_env = cfg.env_height * cfg.env_width                                          #For MLP, will be fully connected graph, so env input shape is the total area
    act_input = Input(shape=(1,), name='Action')
    env_input = Input(shape=(shape_env,), name='Environment')

    #Concatenate Action and Env info
    x = concatenate([act_input, env_input])

    #Create Hidden Layers, i.e. internal neurons
    for i in range(cfg.layer_cnt):
        shape_env=int(shape_env/2)
        x = Dense(units=shape_env, activation='tanh')(x)                                            #Final (x) stacks internal hidden layers on input layers
        x = Dropout(cfg.dropout_rate, noise_shape=None, seed=None)(x)                               #Can experiment with Dropout rate setting                                

    #Create output
    main_output = Dense(units=1, name='Emotion')(x)

    #Create Model (from input, hidden, output
    model = Model(inputs=[act_input, env_input], outputs=[main_output])


    #Compile Model
    model.compile(optimizer='rmsprop',
                  loss='mse',
                  metrics=['mse', 'mae'])

    return model


def MLP_update(model, labels, action_data, env_data):
    '''Learning update function for Basic MLP feed forward neural network'''

    #Reshape env_data, if needed (flatten)       
    env_data2=[]
    for i in range(len(env_data)):
        for j in range(len(env_data[0])):
            env_data2.append(env_data[i][j])
           
    #env_data=env_data2
    
    #Convert training data into numpy array
    shape_env = cfg.env_height * cfg.env_width 
    label=np.asarray([labels]).reshape(1,1)                                                         #Reshape is necessary here, since first dim are "rows" [], and second dim are "columns" of input data [[]]
    action_data=np.asarray([action_data]).reshape(1,1)      
    env_data=np.asarray(env_data).reshape(1,shape_env)

    #Run learning
    results = model.fit([action_data, env_data], [labels], epochs=1, batch_size=1, verbose=0)

    return results.history['loss'][0]                                                               #Since we are just training one epoch at a time, only 1 value in history array, so get that first value out


def MLP_predict(model, action_data, env_data):
    '''Predict function for Basic MLP feed forward neural network'''

    #Reshape env_data, if needed (flatten)      
    env_data2=[[] for i in range(len(action_data))]
    for i in range(len(env_data)):
        for j in range(len(env_data[0])):
            for act in range(len(action_data)):
               env_data2[act].append(env_data[i][j])                                                #For predicting, we need to duplicate env array for each action, to pass into Keras
    env_data=env_data2

    
    #Convert data into numpy array                                                        
    shape_env = cfg.env_height * cfg.env_width 
    action_data=np.asarray(action_data).reshape(len(action_data),1)                                 #Reshape is necessary here, since first dim are "rows" [], and second dim are "columns" of input data [[]]
    env_data=np.asarray(env_data).reshape(len(action_data),shape_env)

    #print(action_data)
    #print(env_data)

    #Make prediction
    pred = model.predict([action_data, env_data], batch_size=len(action_data))

    return pred


#############################################################################
#
# Recurrent Neural Network
# 
##########################

def RNN_construct():
    '''Constructor for RNN neural network'''

    #Create Input Layers
    shape_env = cfg.env_height * cfg.env_width                                          #For RNN, will be fully connected graph, so env input shape is the total area
    act_input = Input(shape=(1,), name='Action')
    env_input = Input(shape=(2,shape_env), name='Environment')                          #Third dimension here necessary for RNN (i.e. 2nd dim in shape) is the time dim that controls degree of recurrence

    #Recurrent step on Env
    rnn = SimpleRNN(units=36)(env_input)
    lstm = LSTM(units=36)(env_input)
    x = lstm

    #Concatenate Action and Env info
    x = concatenate([act_input, x])

    #Create Hidden Layers, i.e. internal neurons
    for i in range(cfg.layer_cnt):
        shape_env=int(shape_env/2)
        x = Dense(units=shape_env, activation='tanh')(x)                                            #Final (x) stacks internal hidden layers on input layers
        x = Dropout(cfg.dropout_rate, noise_shape=None, seed=None)(x)                               #Can experiment with Dropout rate setting                                

    #Create output
    main_output = Dense(units=1, name='Emotion')(x)

    #Create Model (from input, hidden, output
    model = Model(inputs=[act_input, env_input], outputs=[main_output])


    #Compile Model
    model.compile(optimizer='rmsprop',
                  loss='mse',
                  metrics=['mse', 'mae'])

    return model


def RNN_update(model, labels, action_data, env_data):
    '''Learning update function for RNN neural network'''

    #Reshape env_data, if needed (flatten)       
    env_data2=[[] for t in range(2)]
    for t in range(2):
        for i in range(len(env_data[0])):
            for j in range(len(env_data[0][0])):
                env_data2[t].append(env_data[t][i][j])

    env_data=env_data2

    
    #Convert training data into numpy array
    shape_env = cfg.env_height * cfg.env_width 
    label=np.asarray([labels]).reshape(1,1)                                                         #Reshape is necessary here, since first dim are "rows" [], and second dim are "columns" of input data [[]]
    action_data=np.asarray([action_data]).reshape(1,1)      
    env_data=np.asarray(env_data).reshape(1,2,shape_env)                                            #Third dimension here necessary for RNN, because of time dim

    #Run learning
    results = model.fit([action_data, env_data], [labels], epochs=1, batch_size=1, verbose=0)

    return results.history['loss'][0]                                                               #Since we are just training one epoch at a time, only 1 value in history array, so get that first value out


def RNN_predict(model, action_data, env_data):
    '''Predict function for RNN neural network'''

    #Reshape env_data, if needed (flatten)      
    env_data2=[[[] for t in range(2)] for i in range(len(action_data))]
    for act in range(len(action_data)):
        for t in range(2):
            for i in range(len(env_data[0])):
                for j in range(len(env_data[0][0])):
                    env_data2[act][t].append(env_data[t][i][j])                                     #For predicting, we need to duplicate env array for each action, to pass into Keras
                                                                        
    env_data=env_data2
 
    
    #Convert data into numpy array                                                        
    shape_env = cfg.env_height * cfg.env_width 
    action_data=np.asarray(action_data).reshape(len(action_data),1)                                 #Reshape is necessary here, since first dim are "rows" [], and second dim are "columns" of input data [[]]
    env_data=np.asarray(env_data).reshape(len(action_data),2,shape_env)                             #Third dimension here necessary for RNN, because of time dim

    #print(action_data)
    #print(env_data)

    #Make prediction
    pred = model.predict([action_data, env_data], batch_size=len(action_data))

    return pred


